Imports System.Text
Imports System.Threading.Tasks

Namespace KNN
	Friend Class KNNProgram
		Shared Sub Main(ByVal args() As String)
            Console.WriteLine(vbCrLf & "Begin k-NN classification demo" & vbCrLf)

            Dim trainData()() As Double = LoadData() ' get the normalized data

            Dim numFeatures = 2 ' predictor variables
            Dim numClasses = 3 ' 0, 1, 2

            Dim unknown() As Double = {5.25, 1.75}
            Console.WriteLine("Classifying item with predictor value" & "" & "s: 5.25 1.75 " & vbCrLf)

            Dim k = 1
            Console.WriteLine("With k = 1")
            Dim predicted = Classify(unknown, trainData, numClasses, k)
            Console.WriteLine(vbCrLf & "Predicted class = " & predicted)
            Console.WriteLine("")

            k = 4
            Console.WriteLine("With k = 4")
            predicted = Classify(unknown, trainData, numClasses, k)
            Console.WriteLine(vbCrLf & "Predicted class = " & predicted)
            Console.WriteLine("")

            Console.WriteLine("End k-NN demo " & vbCrLf)
            Console.ReadLine()
        End Sub ' Main

        Private Shared Function Classify(ByVal unknown() As Double, ByVal trainData()() As Double, ByVal numClasses As Integer, ByVal k As Integer) As Integer
            ' compute and store distances from unknown to all train data 
            Dim n = trainData.Length ' number data items
            Dim info(n - 1) As IndexAndDistance
            For i = 0 To n - 1
                Dim curr As New IndexAndDistance()
                Dim dist As Double = Distance(unknown, trainData(i))
                curr.idx = i
                curr.dist = dist
                info(i) = curr
            Next i

            Array.Sort(info) ' sort by distance
            Console.WriteLine(vbCrLf & "Nearest  /  Distance  / Class")
            Console.WriteLine("==============================")
            For i = 0 To k - 1
                Dim c = CInt(Fix(trainData(info(i).idx)(2)))
                Dim dist As String = info(i).dist.ToString("F3")
                Console.WriteLine("( " & trainData(info(i).idx)(0) & "," & trainData(info(i).idx)(1) & " )  :  " & dist & "        " & c)
            Next i

            Dim result = Vote(info, trainData, numClasses, k) ' k nearest classes
            Return result

		End Function ' Classify

		Private Shared Function Vote(ByVal info() As IndexAndDistance, ByVal trainData()() As Double, ByVal numClasses As Integer, ByVal k As Integer) As Integer
			Dim votes(numClasses - 1) As Integer ' one cell per class
            For i = 0 To k - 1 ' just first k nearest
                Dim idx = info(i).idx ' which item
                Dim c = CInt(Fix(trainData(idx)(2))) ' class in last cell
                votes(c) += 1
            Next i

            'for (int p = 0; p < votes.Length; ++p)
            '{
            '  Console.Write(votes[p] + " ");
            '}
            'Console.WriteLine("");

            Dim mostVotes = 0
            Dim classWithMostVotes = 0
            For j = 0 To numClasses - 1
                If votes(j) > mostVotes Then
                    mostVotes = votes(j)
                    classWithMostVotes = j
                End If
            Next j

            Return classWithMostVotes
		End Function

		Private Shared Function Distance(ByVal unknown() As Double, ByVal data() As Double) As Double
			Dim sum As Double = 0.0
            For i = 0 To unknown.Length - 1
                sum += (unknown(i) - data(i)) * (unknown(i) - data(i))
            Next i
            Return Math.Sqrt(sum)
		End Function

		Private Shared Function LoadData() As Double()()
			Dim data(32)() As Double
            data(0) = New Double() {2.0, 4.0, 0}
            data(1) = New Double() { 2.0, 5.0, 0 }
			data(2) = New Double() { 2.0, 6.0, 0 }
			data(3) = New Double() { 3.0, 3.0, 0 }
			data(4) = New Double() { 3.0, 7.0, 0 }
            data(5) = New Double() {4.0, 3.0, 0}
            data(6) = New Double() { 4.0, 7.0, 0 }
			data(7) = New Double() { 5.0, 3.0, 0 }
			data(8) = New Double() { 5.0, 7.0, 0 }
			data(9) = New Double() { 6.0, 4.0, 0 }
			data(10) = New Double() { 6.0, 5.0, 0 }
			data(11) = New Double() { 6.0, 6.0, 0 }

			data(12) = New Double() { 3.0, 4.0, 1 }
			data(13) = New Double() { 3.0, 5.0, 1 }
			data(14) = New Double() { 3.0, 6.0, 1 }
			data(15) = New Double() { 4.0, 4.0, 1 }
			data(16) = New Double() { 4.0, 5.0, 1 }
			data(17) = New Double() { 4.0, 6.0, 1 }
			data(18) = New Double() { 5.0, 4.0, 1 }
			data(19) = New Double() { 5.0, 5.0, 1 }
			data(20) = New Double() { 5.0, 6.0, 1 }
			data(21) = New Double() { 6.0, 1.0, 1 }
			data(22) = New Double() { 7.0, 2.0, 1 }
			data(23) = New Double() { 8.0, 3.0, 1 }
			data(24) = New Double() { 8.0, 2.0, 1 }
			data(25) = New Double() { 8.0, 1.0, 1 }
			data(26) = New Double() { 7.0, 1.0, 1 }

			data(27) = New Double() { 2.0, 1.0, 2 }
			data(28) = New Double() { 2.0, 2.0, 2 }
			data(29) = New Double() { 3.0, 1.0, 2 }
			data(30) = New Double() { 3.0, 2.0, 2 }
			data(31) = New Double() { 4.0, 1.0, 2 }
			data(32) = New Double() { 4.0, 2.0, 2 }

			Return data
		End Function

	End Class ' Program

	Public Class IndexAndDistance
		Implements IComparable(Of IndexAndDistance)
        Public idx ' index of a training item
        Public dist As Double ' distance from train item to unknown

		' need to sort these to find k closest
		Public Function CompareTo(ByVal other As IndexAndDistance) As Integer Implements IComparable(Of IndexAndDistance).CompareTo
			If Me.dist < other.dist Then
				Return -1
			ElseIf Me.dist > other.dist Then
				Return +1
			Else
				Return 0
			End If
		End Function
	End Class
End Namespace
